package main

import "fmt"

type (
	JalaliDate struct {
		Day   int
		Month int
		Year  int
	}
	Time struct {
		JalaliDate
		Hour   int
		Minute int
		Second int
	}
	TimePeriod struct {
		Start Time
		End   Time
	}

	Coordinate struct {
		X int
		Y int
	}
	Area struct {
		UpLeft    Coordinate
		DownRight Coordinate
	}

	Discount struct {
		Percent        uint64
		EffectiveArea  Area
		IsActive       bool
		ApplicableTime TimePeriod
	}
	Discounts map[string]*Discount

	SnappSystem struct {
		AllDiscounts        Discounts
		CurrentTime         Time
		BasePriceCalculatur func(TripRequest) Price
	}

	Price uint64

	TripRequest struct {
		From         Coordinate
		To           Coordinate
		basePrice    Price
		UserDiscount string
		RequestTime  Time
	}
)

func (system SnappSystem) findDiscountByName(userDiscount string) (*Discount, error) {
	discount, exists := system.AllDiscounts[userDiscount]
	if !exists {
		return nil, fmt.Errorf("discount [%s] doesn't exists", userDiscount)
	}
	return discount, nil
}

func (area Area) isPointInArea(point Coordinate) bool {
	return point.X <= area.DownRight.X &&
		area.UpLeft.X <= point.X &&
		point.Y <= area.DownRight.Y &&
		area.UpLeft.Y <= point.Y
}

func (base Time) IsAfter(time Time) bool {
	// faster response for common cases
	absoloutelyAfter := base.Year >= time.Year &&
		base.Month >= time.Month &&
		base.Day > time.Day
	if absoloutelyAfter {
		return true
	}

	yEq := base.Year == time.Year
	mEq := base.Month == time.Month
	dEq := base.Day == time.Day

	yAfter := base.Year > time.Year
	mAfter := base.Month > time.Month
	dAfter := base.Day > time.Day

	if yAfter {
		return true
	} else if yEq && mAfter {
		return true
	} else if yEq && mEq && dAfter {
		return true
	} else if yEq && mEq && dEq {
		if base.Hour > time.Hour {
			return true
		} else if base.Hour == time.Hour && base.Minute > time.Minute {
			return true
		} else if base.Hour == time.Hour &&
			base.Minute == time.Minute &&
			base.Second > time.Second {
			return true
		} else {
			return false
		}
	} else {
		return false
	}
}

func (period TimePeriod) isTimeInPeriod(t Time) bool {
	// make the comparison inclusive
	period.Start.Second -= 1
	period.End.Second += 1
	return period.End.IsAfter(t) && t.IsAfter(period.Start)
}

func (disc Discount) Apply(basePrice Price) Price {
	return Price(((100 - disc.Percent) * uint64(basePrice)) / 100)
}

func (system SnappSystem) CalculateEffectivePrice(trip TripRequest) (Price, error) {

	discount, err := system.findDiscountByName(trip.UserDiscount)
	if err != nil {
		return 0, err
	}

	if !discount.EffectiveArea.isPointInArea(trip.From) {
		return 0, fmt.Errorf("user is not in discount area")
	}

	if !discount.IsActive {
		return 0, fmt.Errorf("discount is not currently active")
	}

	if !discount.ApplicableTime.isTimeInPeriod(trip.RequestTime) {
		return 0, fmt.Errorf("discount is not applicable for this time")
	}

	if !system.CurrentTime.IsAfter(trip.RequestTime) {
		return 0, fmt.Errorf("trip time can't be in the future")
	}

	trip.basePrice = system.BasePriceCalculatur(trip)
	finalPrice := discount.Apply(trip.basePrice)

	return finalPrice, nil
}
